# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Maps channels and resolutions to bitrate, size, and profile."""

class ChannelData(object):

  def __init__(self, aac_bitrate, opus_bitrate):
    self.aac_bitrate = aac_bitrate
    self.opus_bitrate = opus_bitrate

class ResolutionData(object):

  def __init__(self, width, height, h264_bitrate, vp9_bitrate, h264_profile):
    self.width = width
    self.height = height
    self.h264_bitrate = h264_bitrate
    self.vp9_bitrate = vp9_bitrate
    self.h264_profile = h264_profile

  def __eq__(self, other):
    return self.height == other.height

  def __ne__(self, other):
    return not self.__eq__(other)

  def __ge__(self, other):
    return self.height >= other.height

# A map of channels to ChannelData objects which contains the AAC and Opus
# bitrate information of a given channel.
# TODO(joeyparrish): Break this down into a multi-level map involving codec.
CHANNEL_MAP = {
    2: ChannelData('128k', '64k'),
    6: ChannelData('192k', '96k'),
}

# A map of resolutions to ResolutionData objects which contain
# the height and bitrate of a given resolution.
# TODO(joeyparrish): Break this down into a multi-level map involving codec.
RESOLUTION_MAP = {
    '144p': ResolutionData(256, 144, '108k', '95k', 'baseline'),
    '240p': ResolutionData(426, 240, '242k', '150k', 'main'),
    '360p': ResolutionData(640, 360, '400k', '276k', 'main'),
    '480p': ResolutionData(854, 480, '2M', '750k', 'main'),
    '576p': ResolutionData(1024, 576, '2.5M', '1M', 'main'),
    '720p': ResolutionData(1280, 720, '3M', '2M', 'main'),
    '720p-hfr': ResolutionData(1280, 720, '4M', '4M', 'main'),
    '1080p': ResolutionData(1920, 1080, '5M', '4M', 'high'),
    '1080p-hfr': ResolutionData(1920, 1080, '6M', '6M', 'high'),
    '2k': ResolutionData(2560, 1440, '9M', '6M', 'high'),
    '2k-hfr': ResolutionData(2560, 1440, '14M', '9M', 'high'),
    '4k': ResolutionData(3840, 2160, '17M', '12M', 'uhd'),
    '4k-hfr': ResolutionData(3840, 2160, '25M', '18M', 'uhd'),
}

class Metadata(object):
  def __init__(self, pipe, channels=None, resolution_name=None,
               codec=None, language=None, hardware=False):
    self.pipe = pipe
    self.channels = channels
    self.codec = codec
    self.language = language
    self.resolution_name = resolution_name
    self.hardware = hardware

    # TODO(joeyparrish): Use constants for codec & format names
    if channels:
      self.type = 'audio'
      # TODO(joeyparrish): channel_data is a weak variable name.
      self.channel_data = CHANNEL_MAP[channels]
      if self.codec == 'aac':
        self.bitrate = self.channel_data.aac_bitrate
        self.format = 'mp4'
      elif self.codec == 'opus':
        self.bitrate = self.channel_data.opus_bitrate
        self.format = 'webm'

    if resolution_name:
      self.type = 'video'
      # TODO(joeyparrish): resolution_data is a weak variable name.
      self.resolution_data = RESOLUTION_MAP[resolution_name]
      if self.codec == 'h264':
        self.bitrate = self.resolution_data.h264_bitrate
        self.format = 'mp4'
      elif self.codec == 'vp9':
        self.bitrate = self.resolution_data.vp9_bitrate
        self.format = 'webm'

  def fill_template(self, template, **kwargs):
    """Fill in a template string using **kwargs and values in self."""
    value_map = {}
    # First take any values from this object itself.
    value_map.update(self.__dict__)
    # Then fill in any values from kwargs.
    value_map.update(kwargs)
    # Now fill in the template with these values.
    return template.format(**value_map)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum
import shlex

from . import configuration
from . import metadata


class InputType(enum.Enum):
  FILE = 'file'
  """A track from a file.  Usable only with VOD."""

  LOOPED_FILE = 'looped_file'
  """A track from a file, looped forever by FFmpeg.  Usable only with live."""

  WEBCAM = 'webcam'
  """A webcam device.  Usable only with live.

  The device path should be given in the name field.  For example, on Linux,
  this might be /dev/video0.
  """

  RAW_IMAGES = 'raw_images'
  """A file or pipe with a sequence of raw images.

  Requires the specification of frame_rate.  May require the use of
  extra_input_args if FFmpeg can't guess the format.
  """

  EXTERNAL_COMMAND = 'external_command'
  """An external command that generates a stream of audio or video.

  The command should be given in the name field, using shell quoting rules.
  The command should send its generated output to the path in the environment
  variable $SHAKA_STREAMER_EXTERNAL_COMMAND_OUTPUT, which Shaka Streamer set to
  the path to the output pipe.

  May require the user of extra_input_args if FFmpeg can't guess the format or
  framerate.
  """

class MediaType(enum.Enum):
  AUDIO = 'audio'
  VIDEO = 'video'
  TEXT = 'text'


# A runtime-created enum for valid resolutions, based on the |metadata| module.
Resolution = configuration.enum_from_keys('Resolution', metadata.RESOLUTION_MAP)


class Input(configuration.Base):
  """An object representing a single input stream to Shaka Streamer."""

  input_type = configuration.Field(InputType, default=InputType.FILE)
  """The type of the input."""

  name = configuration.Field(str, required=True)
  """Name of the input.

  With input_type set to 'file', this is a path to a file name.

  With input_type set to 'looped_file', this is a path to a file name to be
  looped indefinitely in FFmpeg.

  With input_type set to 'webcam', this is a path to the device node for the
  webcam.  For example, on Linux, this might be /dev/video0.

  With input_type set to 'raw_images', this is a path to a file or pipe
  containing a sequence of raw images.

  With input_type set to 'external_command', this is an external command that
  generates a stream of audio or video.  The command will be parsed using shell
  quoting rules.  The command should send its generated output to the path in
  the environment variable $SHAKA_STREAMER_EXTERNAL_COMMAND_OUTPUT, which Shaka
  Streamer set to the path to the output pipe.
  """

  extra_input_args = configuration.Field(str, default='')
  """Extra input arguments needed by FFmpeg to understand the input.

  This allows you to take inputs that cannot be understand or detected
  automatically by FFmpeg.

  This string will be parsed using shell quoting rules.
  """

  media_type = configuration.Field(MediaType, required=True)
  """The media type of the input stream."""

  # TODO: detect frame rate if not given, require only for raw_images
  frame_rate = configuration.Field(float)
  """The frame rate of the input stream, in frames per second.

  Required for media_type of 'video'.
  """

  # TODO: detect resolution if not given
  # TODO: support custom resolutions
  resolution = configuration.Field(Resolution)
  """The name of the input resolution (1080p, etc).

  Required for media_type of 'video'.
  """

  # TODO: detect track number based on media_type
  track_num = configuration.Field(int, default=0)
  """The track number of the input.  Defaults to 0.

  For multiplexed input files, video is usually track 0, and the first audio
  track is usually track 1.  This may vary, though.
  """

  is_interlaced = configuration.Field(bool, default=False)
  """True if the input video is interlaced.

  If true, the video will be deinterlaced during transcoding.
  """

  language = configuration.Field(str)
  """The language of an audio or text stream.

  With input_type set to 'file' or 'looped_file', this will be auto-detected.
  Otherwise, it will default to 'und' (undetermined).
  """

  start_time = configuration.Field(str)
  """The start time of the slice of the input to use.

  Only valid for VOD and with input_type set to 'file'.
  """

  end_time = configuration.Field(str)
  """The end time of the slice of the input to use.

  Only valid for VOD and with input_type set to 'file'.
  """

  filters = configuration.Field(list, subtype=str, default=[])
  """A list of FFmpeg filter strings to add to the transcoding of this input.

  Each filter is a single string.  For example, 'pad=1280:720:20:20'.
  """


  def __init__(self, *args):
    super().__init__(*args)

    if self.media_type == MediaType.VIDEO:
      # These fields are required for video inputs.
      if self.frame_rate is None:
        raise configuration.MissingRequiredField(
            self.__class__, 'frame_rate', self.__class__.frame_rate)

      if self.resolution is None:
        raise configuration.MissingRequiredField(
            self.__class__, 'resolution', self.__class__.resolution)

    if self.input_type != InputType.FILE:
      # These fields are only valid for file inputs.
      reason = 'only valid when input_type is "file"'

      if self.start_time:
        field = self.__class__.start_time
        raise configuration.MalformedField(
            self.__class__, 'start_time', field, reason)

      if self.end_time:
        field = self.__class__.end_time
        raise configuration.MalformedField(
            self.__class__, 'end_time', field, reason)

    # This needs to be parsed into an argument array.  Note that shlex.split on
    # an empty string will produce an empty array.
    self.extra_input_args = shlex.split(self.extra_input_args)

    # A path to a pipe into which this input's contents are fed.
    # None for most input types.
    self._pipe = None

  def set_pipe(self, pipe):
    """Set the path to a pipe into which this input's contents are fed.

    If set, this is what TranscoderNode will read from instead of .name.
    """
    self._pipe = pipe

  def get_path_for_transcode(self):
    """Get the path which the transcoder will use to read the input.

    For some input types, this is a named pipe.  For others, this is .name.
    """
    return self._pipe or self.name


class InputConfig(configuration.Base):
  """An object representing the entire input config to Shaka Streamer."""

  inputs = configuration.Field(list, subtype=Input, required=True)
  """A list of Input objects, one per input stream."""




# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A module that feeds information from two named pipes into shaka-packager."""

import os
import subprocess

from . import metadata
from . import node_base
from . import pipeline_configuration

# Alias a few classes to avoid repeating namespaces later.
ManifestFormat = pipeline_configuration.ManifestFormat
StreamingMode = pipeline_configuration.StreamingMode


INIT_SEGMENT = {
  'audio': {
    'mp4': '{dir}/audio_{language}_{channels}c_{bitrate}_init.mp4',
    'webm': '{dir}/audio_{language}_{channels}c_{bitrate}_init.webm',
  },
  'video': {
    'mp4': '{dir}/video_{resolution_name}_{bitrate}_init.mp4',
    'webm': '{dir}/video_{resolution_name}_{bitrate}_init.webm',
  },
  'text': '{dir}/text_{language}_init.mp4',
}

MEDIA_SEGMENT = {
  'audio': {
    'mp4': '{dir}/audio_{language}_{channels}c_{bitrate}_$Number$.m4s',
    'webm': '{dir}/audio_{language}_{channels}c_{bitrate}_$Number$.webm',
  },
  'video': {
    'mp4': '{dir}/video_{resolution_name}_{bitrate}_$Number$.m4s',
    'webm': '{dir}/video_{resolution_name}_{bitrate}_$Number$.webm',
  },
  'text': '{dir}/text_{language}_$Number$.m4s',
}

SINGLE_SEGMENT = {
  'audio': {
    'mp4': '{dir}/audio_{language}_{channels}c_{bitrate}_output.mp4',
    'webm': '{dir}/audio_{language}_{channels}c_{bitrate}_output.webm',
  },
  'video': {
    'mp4': '{dir}/video_{resolution_name}_{bitrate}_output.mp4',
    'webm': '{dir}/video_{resolution_name}_{bitrate}_output.webm',
  },
  'text': '{dir}/text_{language}_output.mp4',
}

class SegmentError(Exception):
  """Raise when segment is incompatible with format."""
  pass

class PackagerNode(node_base.NodeBase):

  def __init__(self, audio_inputs, video_inputs, text_inputs, output_dir,
               pipeline_config):
    super().__init__()
    self._audio_inputs = audio_inputs
    self._video_inputs = video_inputs
    self._text_inputs = text_inputs
    self._output_dir = output_dir
    self._segment_dir = os.path.join(output_dir, pipeline_config.segment_folder)
    self._pipeline_config = pipeline_config

  def start(self):
    args = [
        'packager',
    ]

    for input in self._audio_inputs:
      dict = {'in': input.pipe, 'stream': 'audio'}
      if input.language != 'und':
        dict['language'] = input.language
      args += self._create_audio_or_video(dict, input)

    for input in self._video_inputs:
      dict = {'in': input.pipe, 'stream': 'video'}
      args += self._create_audio_or_video(dict, input)

    for input in self._text_inputs:
      dict = {
          'in': input.name,
          'stream': 'text',
          'language': input.language,
      }
      args += self._create_text(dict, input.language)

    args += [
        # Segment duration given in seconds.
        '--segment_duration', str(self._pipeline_config.segment_size),
    ]

    if self._pipeline_config.streaming_mode == StreamingMode.LIVE:
      args += [
          # Number of seconds the user can rewind through backwards.
          '--time_shift_buffer_depth',
          str(self._pipeline_config.availability_window),
          # Number of segments preserved outside the current live window.
          '--preserved_segments_outside_live_window', '1',
          # Number of seconds of content encoded/packaged that is ahead of the
          # live edge.
          '--suggested_presentation_delay',
          str(self._pipeline_config.presentation_delay),
          # Number of seconds between manifest updates.
          '--minimum_update_period',
          str(self._pipeline_config.update_period),
      ]

    args += self._setup_manifest_format()

    if self._pipeline_config.encryption.enable:
      args += self._setup_encryption()

    stdout = None
    if self._pipeline_config.debug_logs:
      # Log by writing all Packager output to a file.  Unlike the logging
      # system in ffmpeg, this will stop any Packager output from getting to
      # the screen.
      stdout = open('PackagerNode.log', 'w')

    self._process = self._create_process(args,
                                         stderr=subprocess.STDOUT,
                                         stdout=stdout)

  def _create_text(self, dict, language):
    # TODO: Format using text Metadata objects, which don't exist yet
    # TODO: Generalize and combine with _create_audio_or_video
    if self._pipeline_config.segment_per_file:
      dict['init_segment'] = INIT_SEGMENT['text'].format(
          dir=self._segment_dir, language=language)
      dict['segment_template'] = MEDIA_SEGMENT['text'].format(
          dir=self._segment_dir, language=language)
    else:
      dict['output'] = SINGLE_SEGMENT['text'].format(
          dir=self._segment_dir, language=language)
    return [_packager_stream_arg(dict)]

  def _create_audio_or_video(self, dict, input):
    if self._pipeline_config.segment_per_file:
      dict['init_segment'] = input.fill_template(
          INIT_SEGMENT[input.type][input.format], dir=self._segment_dir)
      dict['segment_template'] = input.fill_template(
          MEDIA_SEGMENT[input.type][input.format], dir=self._segment_dir)
    else:
      dict['output'] = input.fill_template(
          SINGLE_SEGMENT[input.type][input.format], dir=self._segment_dir)

    return [_packager_stream_arg(dict)]

  def _setup_manifest_format(self):
    args = []
    if ManifestFormat.DASH in self._pipeline_config.manifest_format:
      if self._pipeline_config.streaming_mode == StreamingMode.VOD:
        args += [
            '--generate_static_mpd',
        ]
      args += [
          # Generate DASH manifest file.
          '--mpd_output',
          os.path.join(self._output_dir, self._pipeline_config.dash_output),
      ]
    if ManifestFormat.HLS in self._pipeline_config.manifest_format:
      if self._pipeline_config.streaming_mode == StreamingMode.LIVE:
        args += [
            '--hls_playlist_type', 'LIVE',
        ]
      else:
        args += [
            '--hls_playlist_type', 'VOD',
        ]
      args += [
          # Generate HLS playlist file(s).
          '--hls_master_playlist_output',
          os.path.join(self._output_dir, self._pipeline_config.hls_output),
      ]
    return args

  def _setup_encryption(self):
    # Sets up encryption of content.
    args = [
      '--enable_widevine_encryption',
      '--key_server_url', self._pipeline_config.encryption.key_server_url,
      '--content_id', self._pipeline_config.encryption.content_id,
      '--signer', self._pipeline_config.encryption.signer,
      '--aes_signing_key', self._pipeline_config.encryption.signing_key,
      '--aes_signing_iv', self._pipeline_config.encryption.signing_iv,
      '--protection_scheme',
      self._pipeline_config.encryption.protection_scheme.value,
      '--clear_lead', str(self._pipeline_config.encryption.clear_lead),
    ]
    return args


def _packager_stream_arg(opts):
  ret = ''
  for key, value in opts.items():
    ret += key + '=' + value + ','
  return ret




# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import enum
import os
import shlex

from . import configuration
from . import metadata


# A runtime-created enum for valid resolutions, based on the |metadata| module.
Resolution = configuration.enum_from_keys('Resolution', metadata.RESOLUTION_MAP)

# A randomly-chosen content ID in hex.
RANDOM_CONTENT_ID = base64.b16encode(os.urandom(16)).decode('UTF-8')

# The Widevine UAT server URL.
UAT_SERVER = 'https://license.uat.widevine.com/cenc/getcontentkey/widevine_test'

# Credentials for the Widevine test account.
WIDEVINE_TEST_ACCOUNT = 'widevine_test'
WIDEVINE_TEST_SIGNING_KEY = '1ae8ccd0e7985cc0b6203a55855a1034afc252980e970ca90e5202689f947ab9'
WIDEVINE_TEST_SIGNING_IV = 'd58ce954203b7c9a9a9d467f59839249'


class StreamingMode(enum.Enum):
  LIVE = 'live'
  """Indicates a live stream, which has no end."""

  VOD = 'vod'
  """Indicates a video-on-demand (VOD) stream, which is finite."""

class AudioCodec(enum.Enum):
  AAC = 'aac'
  OPUS = 'opus'

# TODO: ideally, we wouldn't have to explicitly list hw: variants
class VideoCodec(enum.Enum):
  H264 = 'h264'
  """H264, also known as AVC."""

  HARDWARE_H264 = 'hw:h264'
  """H264 with hardware encoding."""

  VP9 = 'vp9'
  """VP9."""

  HARDWARE_VP9 = 'hw:vp9'
  """VP9 with hardware encoding."""

class ManifestFormat(enum.Enum):
  DASH = 'dash'
  HLS = 'hls'

class ProtectionScheme(enum.Enum):
  CENC = 'cenc'
  """AES-128-CTR mode."""

  CBCS = 'cbcs'
  """AES-128-CBC mode with pattern encryption."""


class EncryptionConfig(configuration.Base):
  """An object representing the encryption config for Shaka Streamer."""

  enable = configuration.Field(bool, default=False)
  """If true, encryption is enabled.

  Otherwise, all other encryption settings are ignored.
  """

  content_id = configuration.Field(
      configuration.HexString, default=RANDOM_CONTENT_ID)
  """The content ID, in hex.

  If omitted, a random content ID will be chosen for you.
  """

  key_server_url = configuration.Field(str, default=UAT_SERVER)
  """The URL of your key server.

  This is used to generate an encryption key.  By default, it is Widevine's UAT
  server.
  """

  signer = configuration.Field(str, default=WIDEVINE_TEST_ACCOUNT)
  """The name of the signer when authenticating to the key server.

  Defaults to the Widevine test account.
  """

  signing_key = configuration.Field(
      configuration.HexString, default=WIDEVINE_TEST_SIGNING_KEY)
  """The signing key, in hex, when authenticating to the key server.

  Defaults to the Widevine test account's key.
  """

  signing_iv = configuration.Field(
      configuration.HexString, default=WIDEVINE_TEST_SIGNING_IV)
  """The signing IV, in hex, when authenticating to the key server.

  Defaults to the Widevine test account's IV.
  """

  protection_scheme = configuration.Field(ProtectionScheme,
                                          default=ProtectionScheme.CENC)
  """The protection scheme (cenc or cbcs) to use when encrypting."""

  clear_lead = configuration.Field(int, default=10)
  """The seconds of unencrypted media at the beginning of the stream."""


class PipelineConfig(configuration.Base):
  """An object representing the entire pipeline config for Shaka Streamer."""

  streaming_mode = configuration.Field(StreamingMode, required=True)
  """The streaming mode, which can be either 'vod' or 'live'."""

  quiet = configuration.Field(bool, default=False)
  """If true, reduce the level of output.

  Only errors will be shown in quiet mode.
  """

  debug_logs = configuration.Field(bool, default=False)
  """If true, output simple log files from each node.

  No control is given over log filenames.  Logs are written to the current
  working directory.  We do not yet support log rotation.  This is meant only
  for debugging.
  """

  resolutions = configuration.Field(list, subtype=Resolution,
                                    default=[
                                        Resolution._720p,
                                        Resolution._480p,
                                    ])
  """A list of resolution names to encode.

  Any resolution greater than the input resolution will be ignored, to avoid
  upscaling the content.  By default, will encode in 480p and 720p.
  """

  # TODO(joeyparrish): Default to whatever is in the input.
  channels = configuration.Field(int, default=2)
  """The number of audio channels to encode."""

  audio_codecs = configuration.Field(list, subtype=AudioCodec,
                                     default=[AudioCodec.AAC])
  """The audio codecs to encode with."""

  video_codecs = configuration.Field(list, subtype=VideoCodec,
                                     default=[VideoCodec.H264])
  """The video codecs to encode with.

  Note that the prefix "hw:" indicates that a hardware encoder should be
  used.
  """

  manifest_format = configuration.Field(list, subtype=ManifestFormat,
                                        default=[
                                            ManifestFormat.DASH,
                                            ManifestFormat.HLS,
                                        ])
  """A list of manifest formats (dash or hls) to create.

  By default, this will create both.
  """

  dash_output = configuration.Field(str, default='dash.mpd')
  """Output filename for the DASH manifest, if created."""

  hls_output = configuration.Field(str, default='hls.m3u8')
  """Output filename for the HLS master playlist, if created."""

  segment_folder = configuration.Field(str, default='')
  """Sub-folder for segment output (or blank for none)."""

  segment_size = configuration.Field(float, default=4)
  """The length of each segment in seconds."""

  segment_per_file = configuration.Field(bool, default=True)
  """If true, force each segment to be in a separate file.

  Must be true for live content.
  """

  availability_window = configuration.Field(int, default=300)
  """The number of seconds a segment remains available."""

  presentation_delay = configuration.Field(int, default=30)
  """How far back from the live edge the player should be, in seconds."""

  update_period = configuration.Field(int, default=8)
  """How often the player should fetch a new manifest, in seconds."""

  encryption = configuration.Field(EncryptionConfig,
                                   default=EncryptionConfig({}))
  """Encryption settings."""


  def __init__(self, *args):
    super().__init__(*args)

    if self.streaming_mode == StreamingMode.LIVE and not self.segment_per_file:
      field = self.__class__.segment_per_file
      reason = 'must be true when streaming_mode is "live"'
      raise configuration.MalformedField(
          self.__class__, 'segment_per_file', field, reason)




# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import enum
import re

from . import metadata


def enum_from_keys(name, dictionary):
  """Create an enum dynamically from the keys of a dictionary.

  Used to create enums for config validation based on the hard-coded resolution
  settings in the |metadata| module.
  """

  # This is the functional API of the enum module, which is documented here:
  # https://docs.python.org/3/library/enum.html#functional-api

  # Here the keys match the values, unless the key starts with a numeral.
  # Keys starting with numbers would be unusable as enum keys in Python (for
  # example, "480p").  So these will be prefixed with an underscore.
  enum_keys_and_values = []
  for key in dictionary:
    if key[0] >= '0' and key[0] <= '9':
      enum_keys_and_values.append(('_' + key, key))
    else:
      enum_keys_and_values.append((key, key))

  return enum.Enum(name, enum_keys_and_values)


class ConfigError(Exception):
  """A base class for config errors.

  Each subclass provides a meaningful, human-readable string representation in
  English.
  """

  def __init__(self, class_ref, field_name, field):
    self.class_ref = class_ref
    """A reference to the config class that the error refers to."""

    self.class_name = class_ref.__name__
    """The name of the config class that the error refers to."""

    self.field_name = field_name
    """The name of the field that the error refers to."""

    self.field = field
    """The Field metadata object that the error refers to."""


class UnrecognizedField(ConfigError):
  """An error raised when an unrecognized field is encountered in the input."""

  def __str__(self):
    return '{} contains unrecognized field: {}'.format(
        self.class_name, self.field_name)

class WrongType(ConfigError):
  """An error raised when a field in the input has the wrong type."""

  def __str__(self):
    return 'In {}, {} field requires a {}'.format(
        self.class_name, self.field_name, self.field.get_type_name())

class MissingRequiredField(ConfigError):
  """An error raised when a required field is missing from the input."""

  def __str__(self):
    return '{} is missing a required field: {}, a {}'.format(
        self.class_name, self.field_name, self.field.get_type_name())

class MalformedField(ConfigError):
  """An error raised when a field is malformed."""

  def __init__(self, class_ref, field_name, field, reason):
    super().__init__(class_ref, field_name, field)
    self.reason = reason

  def __str__(self):
    return 'In {}, {} field is malformed: {}'.format(
        self.class_name, self.field_name, self.reason)


class ValidatingType(object):
  """A base wrapper type that validates the input against a limited range.

  Subclasses of this are handled specially by the config system, but this is
  not a real class type that can be constructed.

  Subclasses must implement a static validate() method that takes a value and
  raises TypeError if the input type is wrong or ValueError if it fails
  validation.

  Subclasses must define a "name" attribute with a human-readable name for the
  type.
  """

  @abc.abstractmethod
  def __init__(self):
    pass


class HexString(ValidatingType):
  """A wrapper that can be used in Field() to require a hex string."""

  name = 'hexadecimal string'

  @staticmethod
  def validate(value):
    if type(value) is not str:
      raise TypeError()
    if not re.match(r'^[a-fA-F0-9]+$', value):
      raise ValueError('not a hexadecimal string')


class Field(object):
  """A container for metadata about individual config fields."""

  def __init__(self, type, required=False, subtype=None, default=None):
    """
    Args:
        type (class): The required type for values of this field.
        required (bool): True if this field is required on input.
        subtype (class): The required type inside lists (type=list).
        default: The default value if the field is not specified.
    """
    self.type = type
    self.required = required
    self.subtype = subtype
    self.default = default

  def get_type_name(self):
    """Get a human-readable string for the name of self.type."""
    return Field.get_type_name_static(self.type, self.subtype)

  @staticmethod
  def get_type_name_static(type, subtype):
    """Get a human-readable string for the name of type."""

    # Make these special cases a little more readable.
    if type is str:
      # Call it a string, not a "str".
      return 'string'
    elif type is list:
      # Mention the subtype.
      return 'list of {}'.format(Field.get_type_name_static(subtype, None))
    elif type is None:
      # This is only here to allow generic handling of UnrecognizedField errors.
      return 'None'
    elif issubclass(type, enum.Enum):
      # Get the list of valid options as quoted strings.
      options = [repr(str(member.value)) for member in type]
      return '{} (one of {})'.format(type.__name__, ', '.join(options))
    elif issubclass(type, ValidatingType):
      return type.name

    # Otherwise, return the name of the type.
    return type.__name__


class Base(object):
  """A base class for config objects.

  This will handle all validation, type-checking, defaults, and extraction of
  values from an input dictionary.

  Subclasses must define class-level Field objects defining their fields.
  The base class does the rest.
  """

  def __init__(self, dictionary):
    """Ingests, type-checks, and validates the input dictionary."""

    # Collect all the config fields for this type.
    config_fields = {}
    for key, field in self.__class__.__dict__.items():
      if isinstance(field, Field):
        config_fields[key] = field

    for key, value in dictionary.items():
      field = config_fields.get(key)

      # Look for unrecognized fields in the input.
      if not field:
        raise UnrecognizedField(self.__class__, key, Field(None))

      # Check types on recognized fields.
      value = self._check_and_convert_type(field, key, value)

      # Assign the value to self.
      setattr(self, key, value)

    # Look for missing fields.
    for key, field in config_fields.items():
      if not key in dictionary:
        # If it's required, raise an error.
        if field.required:
          raise MissingRequiredField(self.__class__, key, field)

        # Otherwise, assign a default.
        setattr(self, key, field.default)

  def _check_and_convert_type(self, field, key, value):
    """Check the type of |value| and convert it as necessary.

    Args:
        field (Field): The field definition.
        key (str): The name of the field.
        value: The value to be checked and converted.

    This checks the type of |value| according to |field| and performs any
    necessary conversions to the target type.  This is where any type-specific
    logic or special cases are handled.

    Note that automatic type coercion is avoided.  We wouldn't want a string
    containing the word "False" coerced to boolean True.
    """

    # For fields containing other config objects, specially check and convert
    # them.
    if issubclass(field.type, Base):
      # A config object at this stage should be a dictionary.
      if not isinstance(value, dict):
        raise WrongType(self.__class__, key, field)

      # Let the type of the sub-object validate its contents.
      sub_object = field.type(value)

      # Return this typed object, which will be assigned to self.
      return sub_object

    # For lists, check the type of the value itself, then check the subtype of
    # the list items.
    if field.type is list:
      if not isinstance(value, list):
        raise WrongType(self.__class__, key, field)

      subfield = Field(field.subtype)
      try:
        return [self._check_and_convert_type(subfield, key, v) for v in value]
      except WrongType as e:
        # If conversion raises WrongType on the subfield, raise a WrongType
        # error on this field instead.  For any other error, just re-raise.  If
        # conversion succeeds, replace the original item with the converted
        # one.
        if e.field == subfield:
          raise WrongType(self.__class__, key, field) from None
        else:
          raise

      return value

    # For enums, try to cast the value to the enum type, and raise a WrongType
    # error if this fails.
    if issubclass(field.type, enum.Enum):
      try:
        return field.type(value)
      except ValueError:
        raise WrongType(self.__class__, key, field) from None

    # A float should be permissive and accept an int, as well.
    if field.type is float:
      if not isinstance(value, float) and not isinstance(value, int):
        raise WrongType(self.__class__, key, field)
      return value

    # These aren't true types, but validation classes for specific types of
    # limited input.  Run the validate() method to check the input.
    if issubclass(field.type, ValidatingType):
      try:
        field.type.validate(value)
      except TypeError:
        raise WrongType(self.__class__, key, field)
      except ValueError as e:
        raise MalformedField(self.__class__, key, field, str(e))
      return value

    # For all other types, just do a basic type check and return the original
    # value.
    if not isinstance(value, field.type):
      raise WrongType(self.__class__, key, field)
    return value




# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pushes output from packager to cloud."""

import glob
import os
import subprocess
import time

from . import node_base

# This is the HTTP header "Cache-Control" which will be attached to the Cloud
# Storage blobs uploaded by this tool.  When the browser requests a file from
# Cloud Storage, the server will use this as the "Cache-Control" header it
# returns.
#
# Here "no-store" means that the response must not be stored in a cache, and
# "no-transform" means that the response must not be manipulated in any way
# (including Chrome's data saver features which might want to re-encode
# content).
#
# https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Cache-Control
CACHE_CONTROL_HEADER = 'Cache-Control: no-store, no-transform'

COMMON_GSUTIL_ARGS = [
    'gsutil',
    '-q', # quiet mode: report errors, but not progress
    '-h', CACHE_CONTROL_HEADER, # set the appropriate cache header on uploads
    '-m', # parllelize the operation
    'rsync', # operation to perform
    '-C', # still try to push other files if one fails
    '-r', # recurse into folders
]

class CloudNode(node_base.ThreadedNodeBase):
  def __init__(self, input_dir, bucket_url, temp_dir):
    super().__init__(thread_name='cloud', continue_on_exception=True)
    self._input_dir = input_dir
    self._bucket_url = bucket_url
    self._temp_dir = temp_dir

  def _thread_single_pass(self):
    # With recursive=True, glob's ** will also match the base dir.
    manifest_files = (
        glob.glob(self._input_dir + '/**/*.mpd', recursive=True) +
        glob.glob(self._input_dir + '/**/*.m3u8', recursive=True))

    # The manifest at any moment will reference existing segment files.
    # We must be careful not to upload a manifest that references segments that
    # haven't been uploaded yet.  So first we will capture manifest contents,
    # then upload current segments, then upload the manifest contents we
    # captured.

    for manifest_path in manifest_files:
      # The path within the input dir.
      subdir_path = os.path.relpath(manifest_path, self._input_dir)

      # Capture manifest contents, and retry until the file is non-empty or
      # until the thread is killed.
      contents = b''
      while (not contents and
             self.check_status() == node_base.ProcessStatus.Running):
        time.sleep(0.1)

        with open(manifest_path, 'rb') as f:
          contents = f.read()

      # Now that we have manifest contents, put them into a temp file so that
      # the manifests can be pushed en masse later.
      temp_file_path = os.path.join(self._temp_dir, subdir_path)
      # Create any necessary intermediate folders.
      temp_file_dir_path = os.path.dirname(temp_file_path)
      os.makedirs(temp_file_dir_path, exist_ok=True)
      # Write the temp file.
      with open(temp_file_path, 'wb') as f:
        f.write(contents)

    # Sync all files except manifest files.
    args = COMMON_GSUTIL_ARGS + [
        '-d', # delete remote files that are no longer needed
        '-x', '.*m3u8', # skip m3u8 files, which we'll push separately later
        '-x', '.*mpd', # skip mpd files, which we'll push separately later
        self._input_dir, # local input folder to sync
        self._bucket_url, # destination in cloud storage
    ]
    # NOTE: The -d option above will not result in the files ignored by -x
    # being deleted from the remote storage location.
    subprocess.check_call(args)

    # Sync the temporary copies of the manifest files.
    args = COMMON_GSUTIL_ARGS + [
        '-J', # compress all files in transit, since they are text
        self._temp_dir, # local input folder to sync
        self._bucket_url, # destination in cloud storage
    ]
    subprocess.check_call(args)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A module that pushes input to ffmpeg to transcode into various formats."""

from . import input_configuration
from . import node_base
from . import pipeline_configuration

# Alias a few classes to avoid repeating namespaces later.
InputType = input_configuration.InputType
MediaType = input_configuration.MediaType
StreamingMode = pipeline_configuration.StreamingMode


# For H264, there are different profiles with different required command line
# arguments.
PROFILE_ARGS = {
  'baseline': ['-profile:v', 'baseline', '-level:v', '3.0'],
  'main': ['-profile:v', 'main', '-level:v', '3.1'],
  'high': ['-profile:v', 'high', '-level:v', '4.0'],
  'uhd': ['-profile:v', 'high', '-level:v', '5.1'],
}

class TranscoderNode(node_base.NodeBase):

  def __init__(self, output_audios, output_videos, input_config,
               pipeline_config):
    super().__init__()
    self._output_audios = output_audios
    self._output_videos = output_videos
    self._input_config = input_config
    self._pipeline_config = pipeline_config

  def start(self):
    args = [
        'ffmpeg',
        # Do not prompt for output files that already exist. Since we created
        # the named pipe in advance, it definitely already exists. A prompt
        # would block ffmpeg to wait for user input.
        '-y',
    ]

    if self._pipeline_config.quiet:
      args += [
          # Suppresses all messages except errors.
          # Without this, a status line will be printed by default showing
          # progress and transcoding speed.
          '-loglevel', 'error',
      ]

    if any([output.hardware for output in self._output_videos]):
      args += [
          # Hardware acceleration args.
          '-hwaccel', 'vaapi',
          '-vaapi_device', '/dev/dri/renderD128',
      ]

    for input in self._input_config.inputs:
      # The config file may specify additional args needed for this input.
      # This allows, for example, an external-command-type input to generate
      # almost anything ffmpeg could ingest.
      args += input.extra_input_args

      # These are like hard-coded extra_input_args for certain input types.
      # This means users don't have to know much about FFmpeg options to handle
      # these common cases.
      if input.input_type == InputType.LOOPED_FILE:
        args += [
          # Loop the input forever.
          '-stream_loop', '-1',
          # Read input in real time; don't go above 1x processing speed.
          '-re',
        ]
      elif input.input_type == InputType.RAW_IMAGES:
        args += [
            # Parse the input as a stream of images fed into a pipe.
            '-f', 'image2pipe',
            # Set the frame rate to the one specified in the input config.
            # Note that this is the input framerate for the image2 dexuxer, which
            # is not what the similar '-r' option is meant for.
            '-framerate', str(input.frame_rate),
        ]
      elif input.input_type == InputType.WEBCAM:
        args += [
            # Format the input using the webcam format.
            '-f', 'video4linux2',
        ]

      if self._pipeline_config.streaming_mode == StreamingMode.LIVE:
        args += [
            # A larger queue to buffer input from the pipeline (default is 8).
            # This is in packets, but for raw images, that means frames.  A
            # 720p PPM frame is 2.7MB, and a 1080p PPM is 6.2MB.  The entire
            # queue, when full, must fit into memory.
            '-thread_queue_size', '200',
        ]

      if input.start_time:
        args += [
            # Encode from intended starting time of the VOD input.
            '-ss', input.start_time,
        ]
      if input.end_time:
        args += [
            # Encode until intended ending time of the VOD input.
            '-to', input.end_time,
        ]

      # The input name always comes after the applicable input arguments.
      args += [
          # The input itself.
          '-i', input.get_path_for_transcode(),
      ]

    for i, input in enumerate(self._input_config.inputs):
      map_args = [
          # Map corresponding input stream to output file.
          # The format is "<INPUT FILE NUMBER>:<TRACK NUMBER>", so "i" here is
          # the input file number, and "input.get_track()" is the track number
          # from that input file.  The output stream for this input is implied
          # by where we are in the ffmpeg argument list.
          '-map', '{0}:{1}'.format(i, input.track_num),
      ]

      if input.media_type == MediaType.AUDIO:
        for audio in self._output_audios:
          # Map arguments must be repeated for each output file.
          args += map_args
          args += self._encode_audio(audio, input)

      if input.media_type == MediaType.VIDEO:
        for video in self._output_videos:
          # Map arguments must be repeated for each output file.
          args += map_args
          args += self._encode_video(video, input)

    env = {}
    if self._pipeline_config.debug_logs:
      # Use this environment variable to turn on ffmpeg's logging.  This is
      # independent of the -loglevel switch above.
      env['FFREPORT'] = 'file=TranscoderNode.log:level=32'

    self._process = self._create_process(args, env)

  def _encode_audio(self, audio, input):
    filters = []
    args = [
        # No video encoding for audio.
        '-vn',
        # Set the number of channels to the one specified in the VOD config
        # file.
        '-ac', str(audio.channels),
    ]

    if audio.channels == 6:
      filters += [
        # Work around for https://github.com/google/shaka-packager/issues/598,
        # as seen on https://trac.ffmpeg.org/ticket/6974
        'channelmap=channel_layout=5.1',
      ]

    filters.extend(input.filters)

    if audio.codec == 'aac':
      args += [
          # Format with MPEG-TS for a pipe.
          '-f', 'mpegts',
          # AAC audio codec.
          '-c:a', 'aac',
          # Set bitrate to the one specified in the VOD config file.
          '-b:a', '{0}'.format(audio.channel_data.aac_bitrate),
      ]
    elif audio.codec == 'opus':
      args += [
          # Opus encoding has output format webm.
          '-f', 'webm',
          # Opus audio codec.
          '-c:a', 'libopus',
          # Set bitrate to the one specified in the VOD config file.
          '-b:a', '{0}'.format(audio.channel_data.opus_bitrate),
          # DASH-compatible output format.
          '-dash', '1',
      ]

    if len(filters):
      args += [
          # Set audio filters.
          '-af', ','.join(filters),
      ]

    args += [
        # The output.
        audio.pipe,
    ]
    return args

  # TODO(joeyparrish): "video" is a weak variable name
  def _encode_video(self, video, input):
    filters = []
    args = [
        # No audio encoding for video.
        '-an',
        # Full pelME compare function.
        '-cmp', 'chroma',
    ]

    # TODO: auto detection of interlacing
    if input.is_interlaced:
      frame_rate = input.frame_rate
      # Sanity check: since interlaced files are made up of two interlaced
      # frames, the frame rate must be even and not too small.
      assert frame_rate % 2 == 0 and frame_rate >= 48
      filters.append('pp=fd')
      args.extend(['-r', str(frame_rate / 2)])

    filters.extend(input.filters)

    if video.hardware:
      filters.append('format=nv12')
      filters.append('hwupload')
      # -2 here means to choose a width to keep the original aspect ratio.
      filters.append('scale_vaapi=-2:{0}'.format(video.resolution_data.height))
    else:
      # -2 here means to choose a width to keep the original aspect ratio.
      filters.append('scale=-2:{0}'.format(video.resolution_data.height))

    if video.codec == 'h264':
      args += [
          # MPEG-TS format works well in a pipe.
          '-f', 'mpegts',
      ]

      if self._pipeline_config.streaming_mode == StreamingMode.LIVE:
        args += [
            # Encodes with highest-speed presets for real-time live streaming.
            '-preset', 'ultrafast',
        ]
      else:
        args += [
            # Take your time for VOD streams.
            '-preset', 'slow',
            # Apply the loop filter for higher quality output.
            '-flags', '+loop',
        ]

      if video.hardware:
        args += [
            # H264 VAAPI video codec.
            '-c:v', 'h264_vaapi',
        ]
      else:
        args += [
            # H264 video codec.
            '-c:v', 'h264',
        ]

      args += [
          # Set bitrate to the one specified in the VOD config file.
          '-b:v', '{0}'.format(video.resolution_data.h264_bitrate),
          # Set maximum number of B frames between non-B frames.
          '-bf', '0',
          # The only format supported by QT/Apple.
          '-pix_fmt', 'yuv420p',
          # Require a closed GOP.  Some decoders don't support open GOPs.
          '-flags', '+cgop',
      ]
      # Use different ffmpeg options depending on the H264 profile.
      args += PROFILE_ARGS[video.resolution_data.h264_profile]

    elif video.codec == 'vp9':
      args += [
          # Format using webm.
          '-f', 'webm',
      ]

      if video.hardware:
        args += [
            # VP9 VAAPI video codec.
            '-c:v', 'vp9_vaapi',
        ]
      else:
        args += [
          # VP9 video codec.
          '-c:v', 'vp9',
        ]

      args += [
          # Set bitrate to the one specified in the VOD config file.
          '-b:v', '{0}'.format(video.resolution_data.vp9_bitrate),
          # DASH-compatible output format.
          '-dash', '1',
      ]

    # TODO: auto-detection of framerate?
    keyframe_interval = int(self._pipeline_config.segment_size *
                            input.frame_rate)
    args += [
        # Set minimum and maximum GOP length.
        '-keyint_min', str(keyframe_interval), '-g', str(keyframe_interval),
        # Set video filters.
        '-vf', ','.join(filters),
        # The output.
        video.pipe,
    ]
    return args



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Top-level module API.

If you'd like to import Shaka Streamer as a Python module and build it into
your own application, this is the top-level API you can use for that.  You may
also want to look at the source code to the command-line front end script
`shaka-streamer`.
"""


import os
import re
import shutil
import string
import subprocess
import tempfile
import uuid

from . import cloud_node
from . import external_command_node
from . import input_configuration
from . import metadata
from . import node_base
from . import packager_node
from . import pipeline_configuration
from . import transcoder_node

# Alias a few classes to avoid repeating namespaces later.
InputConfig = input_configuration.InputConfig
InputType = input_configuration.InputType
PipelineConfig = pipeline_configuration.PipelineConfig
StreamingMode = pipeline_configuration.StreamingMode


class ControllerNode(object):
  """Controls all other nodes and manages shared resources."""

  def __init__(self):
    global_temp_dir = tempfile.gettempdir()

    # The docs state that if any of prefix, suffix, or dir are specified, all
    # must be specified (and not None).  Create a temp dir of our own, inside
    # the global temp dir, and with a name that indicates who made it.
    self._temp_dir = tempfile.mkdtemp(
        dir=global_temp_dir, prefix='shaka-live-', suffix='')

    self._nodes = []

  def __del__(self):
    # Clean up named pipes by removing the temp directory we placed them in.
    shutil.rmtree(self._temp_dir)

  def __enter__(self):
    return self

  def __exit__(self, *unused_args):
    self.stop()

  def _create_pipe(self):
    """Create a uniquely-named named pipe in the node's temp directory.

    Raises:
      RuntimeError: If the platform doesn't have mkfifo.
    Returns:
      The path to the named pipe, as a string.
    """

    # TODO(#8): mkfifo only works on Unix.  We would need a special case for a
    # Windows port some day.

    if not hasattr(os, 'mkfifo'):
      raise RuntimeError('Platform not supported due to lack of mkfifo')

    # Since the tempfile module creates actual files, use uuid to generate a
    # filename, then call mkfifo to create the named pipe.
    unique_name = str(uuid.uuid4())
    path = os.path.join(self._temp_dir, unique_name)

    readable_by_owner_only = 0o600  # Unix permission bits
    os.mkfifo(path, mode=readable_by_owner_only)

    return path

  def start(self, output_dir, input_config_dict, pipeline_config_dict,
            bucket_url=None):
    """Create and start all other nodes.

    :raises: `RuntimeError` if the controller has already started.
    :raises: :class:`streamer.configuration.ConfigError` if the configuration is
             invalid.
    """

    if self._nodes:
      raise RuntimeError('Controller already started!')

    # Check that ffmpeg version is 4.1 or above.
    _check_version('FFmpeg', ['ffmpeg', '-version'], (4, 1))

    # Check that Shaka Packager version is 2.1 or above.
    _check_version('Packager', ['packager', '-version'], (2, 1))

    input_config = InputConfig(input_config_dict)
    pipeline_config = PipelineConfig(pipeline_config_dict)
    self._pipeline_config = pipeline_config

    # External command inputs need to be processed by an additional node before
    # being transcoded.  In this case, the input doesn't have a filename that
    # FFmpeg can read, so we generate an intermediate pipe for that node to
    # write to.  TranscoderNode will then instruct FFmpeg to read from that
    # pipe for this input.
    for input in input_config.inputs:
      if input.input_type == InputType.EXTERNAL_COMMAND:
        command_output = self._create_pipe()
        command_node = external_command_node.ExternalCommandNode(
            input.name, command_output)
        self._nodes.append(command_node)
        input.set_pipe(command_output)

    # TODO: This is unnecessary.  Just process each input in turn.
    media_outputs = {
        'audio': [],
        'video': [],
        'text': [],
    }

    # Sorting the media by type.  So all the audio streams are in one list,
    # all the video stream are in one list, etc.
    for media in input_config.inputs:
      media_type = media.media_type.value
      media_outputs[media_type].append(media)

    audio_outputs = []
    for i in media_outputs['audio']:
      audio_outputs.extend(self._add_audio(i,
                                           pipeline_config.channels,
                                           pipeline_config.audio_codecs))

    video_outputs = []
    for i in media_outputs['video']:
      video_outputs.extend(self._add_video(i,
                                           pipeline_config.resolutions,
                                           pipeline_config.video_codecs))

    # Process input through a transcoder node using ffmpeg.
    ffmpeg_node = transcoder_node.TranscoderNode(audio_outputs,
                                                 video_outputs,
                                                 input_config,
                                                 pipeline_config)

    self._nodes.append(ffmpeg_node)

    # Process input through a packager node using Shaka Packager.
    package_node = packager_node.PackagerNode(audio_outputs,
                                              video_outputs,
                                              media_outputs['text'],
                                              output_dir,
                                              pipeline_config)
    self._nodes.append(package_node)

    if bucket_url:
      cloud_temp_dir = os.path.join(self._temp_dir, 'cloud')
      os.mkdir(cloud_temp_dir)

      push_to_cloud = cloud_node.CloudNode(output_dir, bucket_url,
                                           cloud_temp_dir)
      self._nodes.append(push_to_cloud)

    for node in self._nodes:
      node.start()
    return self

  def check_status(self):
    """Checks the status of all the nodes.

    :rtype: streamer.node_base.ProcessStatus

    If one node is errored, this returns Errored; otherwise if one node is
    finished, this returns Finished; this only returns Running if all nodes are
    running.  If there are no nodes, this returns Finished.
    """
    if not self._nodes:
      return node_base.ProcessStatus.Finished

    value = max(node.check_status().value for node in self._nodes)
    return node_base.ProcessStatus(value)

  def stop(self):
    """Stop all nodes."""
    for node in self._nodes:
      node.stop()
    self._nodes = []

  def _add_audio(self, input, channels, codecs):
    audio_outputs = []
    language = input.language or self._probe_language(input)
    for codec in codecs:
      codec = codec.value
      audio_outputs.append(metadata.Metadata(self._create_pipe(),
                                             channels=channels, codec=codec,
                                             language=language))
    return audio_outputs

  def _add_video(self, input, resolutions, codecs):
    video_outputs = []
    for codec in codecs:
      codec = codec.value
      hardware_encoding = False
      if codec.startswith('hw:'):
        hardware_encoding = True
        codec = codec.split(':')[1]

      # TODO: on ingest, convert the resolution string into a value from the map
      in_res = input.resolution.value
      for out_res in resolutions:
        out_res = out_res.value
        # Only going to output lower or equal resolution videos.
        # Upscaling is costly and does not do anything.
        if (metadata.RESOLUTION_MAP[in_res] >=
            metadata.RESOLUTION_MAP[out_res]):
          video_outputs.append(metadata.Metadata(self._create_pipe(),
                                                 resolution_name=out_res,
                                                 codec=codec,
                                                 hardware=hardware_encoding))

    return video_outputs

  # TODO: Move to input_configuration
  def _probe_language(self, input):
    # ffprobe {input}: list out metadata of input
    # -show_entries stream=index:stream_tags=language: list out tracks with
    # stream and language information
    # -select_streams {track}: Only return stream/language information for
    # specified track.
    # -of compact=p=0:nk=1: Specify no keys printed and don't print the name
    # at the beginning of each line.
    command = ['ffprobe', input.name, '-show_entries',
               'stream=index:stream_tags=language', '-select_streams',
               str(input.track_num), '-of', 'compact=p=0:nk=1']

    lang_str = subprocess.check_output(
        command, stderr=subprocess.DEVNULL).decode('utf-8')
    # The regex is looking for a string that is of the format number|language.
    # Once it finds a number| match, it will copy the string until the end of
    # the line.
    lang_match = re.search(r'\d+\|(.*$)', lang_str)
    if lang_match:
      return lang_match.group(1)
    return 'und'

  def is_vod(self):
    """Returns True if the pipeline is running in VOD mode.

    :rtype: bool
    """

    return self._pipeline_config.streaming_mode == StreamingMode.VOD

class VersionError(Exception):
  """A version error for one of Shaka Streamer's external dependencies.

  Raised when a dependency (like FFmpeg) is missing or not new enough to work
  with Shaka Streamer.  See also :doc:`prerequisites`.
  """

  pass

def _check_version(name, command, minimum_version):
  min_version_string = '.'.join([str(x) for x in minimum_version])

  try:
    version_string = str(subprocess.check_output(command))
  except (subprocess.CalledProcessError, OSError) as e:
    if isinstance(e, subprocess.CalledProcessError):
      print(e.stdout, file=sys.stderr)
    raise FileNotFoundError(name + ' not installed! Please install version ' +
                            min_version_string + ' or higher of ' + name + '.')

  version_match = re.search(r'([0-9]+)\.([0-9]+)\.([0-9]+)', version_string)

  if version_match == None:
    raise VersionError(name + ' version not found in string output!')

  version = (int(version_match.group(1)), int(version_match.group(2)))
  if version < minimum_version:
    raise VersionError(name + ' not installed! Please install version ' +
                       min_version_string + ' or higher of ' + name + '.')



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A module that runs an external command to generate media."""

from . import node_base

class ExternalCommandNode(node_base.NodeBase):

  def __init__(self, command, output_path):
    super().__init__()
    self._command = command
    self._output_path = output_path

  def start(self):
    # This environment/shell variable must be used by the external command as
    # the place it sends its generated output.  Since the command is executed
    # with shell=True, the command can include
    # $SHAKA_STREAMER_EXTERNAL_COMMAND_OUTPUT at any point.
    env = {
      'SHAKA_STREAMER_EXTERNAL_COMMAND_OUTPUT': self._output_path,
    }
    # The yaml file may contain a multi-line string, which seems to cause
    # subprocess to execute each line as a command when shell=True.  So convert
    # newlines into spaces.
    command = self._command.replace('\n', ' ')
    self._process = self._create_process(command, shell=True, env=env)
